from __future__ import annotations
from PyQt6.QtOpenGLWidgets import QOpenGLWidget
from PyQt6.QtCore import Qt, QSize, QPoint, QTimer
from PyQt6.QtGui import QColor, QFont, QPainter
from OpenGL.GL import *
from abc import ABC, abstractmethod
import math
from PyQt6.QtWidgets import QApplication
from typing import Callable


class NodeEditorObject():
    def __init__(self):
        self.raycast_hit = False

    def raycast_reset(self, editor: EclipseNodeEditor):
        """Reset the raycast hit"""
        pass

    def raycast(self, editor: EclipseNodeEditor):
        """Handle input for this object"""
        return False

    def input(self, editor: EclipseNodeEditor):
        """Handle input for this object"""
        pass

    def update(self, editor: EclipseNodeEditor):
        """Update object state"""
        pass

    def render(self, editor: EclipseNodeEditor):
        """Render the object"""
        pass


class NodeEditorTool():
    def __init__(self):
        self.is_active = True

    def input(self, editor: EclipseNodeEditor):
        pass

    def update(self, editor: EclipseNodeEditor):
        pass

    def render(self, editor: EclipseNodeEditor):
        pass


class EclipseNodeEditor(QOpenGLWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        # Configure OpenGL format for antialiasing
        format = self.format()
        format.setSamples(4)  # Use 4x multisampling
        self.setFormat(format)

        self.background_color = QColor(21, 21, 21)
        # Set update behavior to redraw continuously
        self.setUpdateBehavior(QOpenGLWidget.UpdateBehavior.NoPartialUpdate)

        # Enable mouse tracking for smooth pan updates
        self.setMouseTracking(True)

        # Enable key tracking
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

        # Initialize view transform (current values)
        self.pan_x = 0
        self.pan_y = 0
        self.zoom = 1.0

        # Target transform values
        self.target_pan_x = 0
        self.target_pan_y = 0
        self.target_zoom = 1.0

        # Transform constraints
        self.min_zoom = 0.1
        self.max_zoom = 10.0

        # Interpolation settings
        self.lerp_factor = 0.2

        # Mouse state
        self.screen_mouse_pos = QPoint()
        self.world_mouse_pos = QPoint()
        self.last_mouse_drag_pos = QPoint()
        self.is_panning = False
        self.left_mouse_down_prev = False
        self.left_mouse_down = False
        self.left_mouse_pressed = False
        self.left_mouse_released = False
        self.right_mouse_down_prev = False
        self.right_mouse_down = False
        self.right_mouse_pressed = False
        self.right_mouse_released = False

        # Keyboard state
        self.keys_down = set()
        self.keys_pressed = set()
        self.keys_released = set()
        self.modifiers_down = set()
        self.modifiers_pressed = set()
        self.modifiers_released = set()

        # Render groups
        self.render_groups = {}

        # Objects list
        self.objects = []
        self.grabbed_object = None

        # Tools list
        self.tools = []
        self.locked_tool = None
        self.tool_unlocking = False
        # QPainter for text rendering
        self.painter = QPainter()

        # Setup update timer for smooth interpolation and object updates
        self.update_timer = QTimer(self)
        self.update_timer.timeout.connect(self.on_timer_update)
        self.update_timer.start(16)  # ~60 FPS

    def add_object(self, obj):
        """Add an object to the editor"""
        self.objects.append(obj)
        return obj

    def remove_object(self, obj):
        """Remove an object from the editor"""
        if obj in self.objects:
            self.objects.remove(obj)

    def get_objects_of_type(self, type: type):
        """Get all objects of a given type"""
        return [obj for obj in self.objects if isinstance(obj, type)]

    def add_tool(self, tool: NodeEditorTool):
        """Add a tool to the editor"""
        self.tools.append(tool)
        return tool

    def remove_tool(self, tool: NodeEditorTool):
        """Remove a tool from the editor"""
        if tool in self.tools:
            self.tools.remove(tool)

    def lock_tool(self, tool: NodeEditorTool):
        """Lock a tool to the editor"""
        self.locked_tool = tool

    def unlock_tool(self):
        """Unlock a tool from the editor"""
        self.tool_unlocking = True
        self.locked_tool = None

    def on_timer_update(self):
        """Handle timer updates for transform interpolation and object updates"""

        # Update current mouse position
        self.screen_mouse_pos = self.mapFromGlobal(self.cursor().pos())

        # Update world mouse position
        world_x, world_y = self.screen_to_world(
            self.screen_mouse_pos.x(), self.screen_mouse_pos.y())
        self.world_mouse_pos.setX(int(world_x))
        self.world_mouse_pos.setY(int(world_y))

        # Update mouse button states
        self.left_mouse_down_prev = self.left_mouse_down
        self.right_mouse_down_prev = self.right_mouse_down
        buttons = QApplication.mouseButtons()
        self.left_mouse_down = bool(buttons & Qt.MouseButton.LeftButton)
        self.right_mouse_down = bool(buttons & Qt.MouseButton.RightButton)
        self.left_mouse_pressed = self.left_mouse_down and not self.left_mouse_down_prev
        self.left_mouse_released = not self.left_mouse_down and self.left_mouse_down_prev
        self.right_mouse_pressed = self.right_mouse_down and not self.right_mouse_down_prev
        self.right_mouse_released = not self.right_mouse_down and self.right_mouse_down_prev

        # Update modifier states
        current_modifiers = QApplication.keyboardModifiers()

        # Track pressed and released states for modifiers
        for modifier in [Qt.KeyboardModifier.ShiftModifier, Qt.KeyboardModifier.ControlModifier, Qt.KeyboardModifier.AltModifier]:
            is_down = bool(current_modifiers & modifier)
            was_down = modifier in self.modifiers_down

            if is_down and not was_down:
                self.modifiers_pressed.add(modifier)
                self.modifiers_down.add(modifier)
            elif not is_down and was_down:
                self.modifiers_released.add(modifier)
                self.modifiers_down.remove(modifier)

        # First handle transform interpolation
        self.interpolate_transform()

        # Then update all objects
        self.process_objects()

        # Clear one-frame states at the END of the frame, after all processing is done
        self.keys_pressed.clear()
        self.keys_released.clear()
        self.modifiers_pressed.clear()
        self.modifiers_released.clear()

    def process_objects(self):
        """Process input and updates for all objects"""
        self.tool_unlocking = False

        # Reset raycast hit
        for obj in self.objects:
            # Raycast hit is more core to the object so we handle that manually
            obj.raycast_hit = False
            obj.raycast_reset(self)

        # Raycast all objects
        for obj in reversed(self.objects):
            if obj.raycast(self):
                obj.raycast_hit = True
                if self.grabbed_object == None and self.left_mouse_down:
                    self.grabbed_object = obj
                break

        if self.left_mouse_released:
            self.grabbed_object = None

        # Process all inputs
        for obj in self.objects:
            obj.input(self)

        # Process all tools
        for tool in self.tools:
            if tool.is_active:
                if self.locked_tool == None:
                    if not self.tool_unlocking:
                        tool.input(self)
                elif self.locked_tool == tool:
                    tool.input(self)

        # Update all objects
        for obj in self.objects:
            obj.update(self)

        # Update all tools
        for tool in self.tools:
            if tool.is_active:
                if self.locked_tool == None:
                    if not self.tool_unlocking:
                        tool.update(self)
                elif self.locked_tool == tool:
                    tool.update(self)

    def get_mouse_pos(self):
        return self.world_mouse_pos.x(), self.world_mouse_pos.y()

    def add_render_group(self, group_index: int, callback: Callable):
        """Adds a callback to a list for this group index"""
        if group_index not in self.render_groups:
            self.render_groups[group_index] = []
        self.render_groups[group_index].append(callback)

    def render_objects(self):
        """Render all objects"""
        # Clear stored render groups
        self.render_groups = {}

        # Call all render methods
        for obj in self.objects:
            obj.render(self)

        # Call all tool render methods
        for tool in self.tools:
            if tool.is_active:
                if self.locked_tool == None:
                    if not self.tool_unlocking:
                        tool.render(self)
                elif self.locked_tool == tool:
                    tool.render(self)

        # Sort render groups and call callbacks in order
        sorted_groups = sorted(self.render_groups.keys())
        for group_index in sorted_groups:
            for callback in self.render_groups[group_index]:
                callback()

    def lerp(self, start, end, factor):
        """Linear interpolation between start and end values"""
        return start + (end - start) * factor

    def interpolate_transform(self):
        """Interpolate between current and target transform values"""
        # Check if we need to update pan
        if abs(self.target_pan_x - self.pan_x) > 0.01 or abs(self.target_pan_y - self.pan_y) > 0.01:
            self.pan_x = self.lerp(
                self.pan_x, self.target_pan_x, self.lerp_factor)
            self.pan_y = self.lerp(
                self.pan_y, self.target_pan_y, self.lerp_factor)

        # Check if we need to update zoom
        if abs(self.target_zoom - self.zoom) > 0.0001:
            self.zoom = self.lerp(
                self.zoom, self.target_zoom, self.lerp_factor)

        # Always update to ensure continuous frame updates
        self.update()

    def initializeGL(self):
        """Initialize OpenGL settings"""
        glClearColor(21/255, 21/255, 21/255, 1.0)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        # Enable multisampling for antialiasing
        glEnable(GL_MULTISAMPLE)
        # Enable line smoothing for better line antialiasing
        glEnable(GL_LINE_SMOOTH)
        glHint(GL_LINE_SMOOTH_HINT, GL_NICEST)

    def draw_rectangle(self, x, y, width, height, r, g, b, a=1.0):
        """Helper method to draw a rectangle"""
        glColor4f(r, g, b, a)
        glBegin(GL_QUADS)
        glVertex2f(x, y)
        glVertex2f(x + width, y)
        glVertex2f(x + width, y + height)
        glVertex2f(x, y + height)
        glEnd()

    def draw_rectangle_outline(self, x, y, width, height, r, g, b, a=1.0, thickness=1.0):
        """Helper method to draw a rectangle outline with specified thickness"""
        glColor4f(r, g, b, a)

        # Draw top line
        glBegin(GL_QUADS)
        glVertex2f(x - thickness/2, y - thickness/2)
        glVertex2f(x + width + thickness/2, y - thickness/2)
        glVertex2f(x + width + thickness/2, y + thickness/2)
        glVertex2f(x - thickness/2, y + thickness/2)
        glEnd()

        # Draw bottom line
        glBegin(GL_QUADS)
        glVertex2f(x - thickness/2, y + height - thickness/2)
        glVertex2f(x + width + thickness/2, y + height - thickness/2)
        glVertex2f(x + width + thickness/2, y + height + thickness/2)
        glVertex2f(x - thickness/2, y + height + thickness/2)
        glEnd()

        # Draw left line
        glBegin(GL_QUADS)
        glVertex2f(x - thickness/2, y + thickness/2)
        glVertex2f(x + thickness/2, y + thickness/2)
        glVertex2f(x + thickness/2, y + height - thickness/2)
        glVertex2f(x - thickness/2, y + height - thickness/2)
        glEnd()

        # Draw right line
        glBegin(GL_QUADS)
        glVertex2f(x + width - thickness/2, y + thickness/2)
        glVertex2f(x + width + thickness/2, y + thickness/2)
        glVertex2f(x + width + thickness/2, y + height - thickness/2)
        glVertex2f(x + width - thickness/2, y + height - thickness/2)
        glEnd()

    def draw_text(self, x, y, text, r, g, b, a=1.0, font_size=12):
        """Helper method to draw text using Qt's painter"""
        # Start QPainter
        self.painter.begin(self)
        self.painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Set up the font
        font = QFont()
        font.setPointSizeF(font_size * self.zoom)
        self.painter.setFont(font)

        # Set color
        self.painter.setPen(
            QColor(int(r * 255), int(g * 255), int(b * 255), int(a * 255)))

        # Convert position to screen space
        screen_x = x * self.zoom + self.pan_x
        screen_y = y * self.zoom + self.pan_y

        # Get font metrics for proper vertical centering
        metrics = self.painter.fontMetrics()
        text_height = metrics.height()

        # Draw the text centered vertically (y is the center point)
        # Subtract half the text height to center it vertically
        self.painter.drawText(int(screen_x), int(
            screen_y - text_height/2 + metrics.ascent()), text)

        # End QPainter
        self.painter.end()

        # Restore OpenGL state
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        self.apply_transform()

    def draw_circle(self, x, y, radius, r, g, b, a=1.0):
        """Helper method to draw a circle"""
        glColor4f(r, g, b, a)
        glBegin(GL_TRIANGLE_FAN)
        glVertex2f(x, y)  # Center point
        segments = 32  # Number of segments to make the circle smooth
        for i in range(segments + 1):
            angle = 2.0 * math.pi * i / segments
            dx = radius * math.cos(angle)
            dy = radius * math.sin(angle)
            glVertex2f(x + dx, y + dy)
        glEnd()

    def draw_line(self, ax, ay, bx, by, thickness,
                  start_r, start_g, start_b, start_a=1.0,
                  end_r=None, end_g=None, end_b=None, end_a=None):
        """Draw a line with gradient between two colors"""
        # Set end color to start color if not specified
        end_r = start_r if end_r is None else end_r
        end_g = start_g if end_g is None else end_g
        end_b = start_b if end_b is None else end_b
        end_a = start_a if end_a is None else end_a

        # Calculate the direction vector of the line
        dx = bx - ax
        dy = by - ay
        length = math.sqrt(dx * dx + dy * dy)

        if length < 0.0001:  # Avoid division by zero for very short lines
            return

        # Calculate the normalized perpendicular vector
        nx = -dy / length * thickness / 2
        ny = dx / length * thickness / 2

        # Draw the quad with color interpolation
        glBegin(GL_QUADS)
        # Start color for first two vertices
        glColor4f(start_r, start_g, start_b, start_a)
        glVertex2f(ax + nx, ay + ny)
        glVertex2f(ax - nx, ay - ny)
        # End color for last two vertices
        glColor4f(end_r, end_g, end_b, end_a)
        glVertex2f(bx - nx, by - ny)
        glVertex2f(bx + nx, by + ny)
        glEnd()

    def screen_to_world(self, screen_x, screen_y):
        """Convert screen coordinates to world coordinates"""
        world_x = (screen_x - self.pan_x) / self.zoom
        world_y = (screen_y - self.pan_y) / self.zoom
        return world_x, world_y

    def world_to_screen(self, world_x, world_y):
        """Convert world coordinates to screen coordinates"""
        screen_x = world_x * self.zoom + self.pan_x
        screen_y = world_y * self.zoom + self.pan_y
        return screen_x, screen_y

    def apply_transform(self):
        """Apply the current view transform"""
        width = self.width()
        height = self.height()

        # Set up orthographic projection
        glLoadIdentity()
        glOrtho(0, width, height, 0, -1, 1)

        # Apply transforms in correct order
        glTranslatef(self.pan_x, self.pan_y, 0)
        glScalef(self.zoom, self.zoom, 1.0)

    def paintGL(self):
        """Paint the scene"""
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        # Apply the view transform
        self.apply_transform()

        # Render all objects
        self.render_objects()

    def resizeGL(self, width, height):
        """Handle window resize"""
        glViewport(0, 0, width, height)

    def mousePressEvent(self, event):
        """Handle mouse press events"""
        if event.button() == Qt.MouseButton.MiddleButton:
            self.is_panning = True
            self.last_mouse_drag_pos = event.pos()
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        """Handle mouse release events"""
        if event.button() == Qt.MouseButton.MiddleButton:
            self.is_panning = False
        super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event):
        """Handle mouse move events"""
        # Update current mouse position
        if self.is_panning:
            delta = event.pos() - self.last_mouse_drag_pos
            self.target_pan_x += delta.x()
            self.target_pan_y += delta.y()
            self.last_mouse_drag_pos = event.pos()
        super().mouseMoveEvent(event)

    def wheelEvent(self, event):
        """Handle mouse wheel events for zooming"""
        # Get mouse position
        mouse_pos = event.position()
        mouse_x = mouse_pos.x()
        mouse_y = mouse_pos.y()

        # Get the world point under mouse before zoom
        old_world_x, old_world_y = self.screen_to_world(mouse_x, mouse_y)

        # Calculate and clamp new zoom level
        zoom_factor = 1.2 if event.angleDelta().y() > 0 else 1/1.2
        new_zoom = self.target_zoom * zoom_factor

        if self.min_zoom <= new_zoom <= self.max_zoom:
            # Set target zoom
            self.target_zoom = new_zoom

            # Calculate the screen point for this world position at target zoom
            # We need to use current pan but target zoom for this calculation
            screen_x = old_world_x * new_zoom + self.target_pan_x
            screen_y = old_world_y * new_zoom + self.target_pan_y

            # Set target pan to keep world point under mouse
            self.target_pan_x += mouse_x - screen_x
            self.target_pan_y += mouse_y - screen_y

        event.accept()

    def keyPressEvent(self, event):
        """Handle key press events"""
        key = event.key()

        # Only handle regular keys here, modifiers are handled in timer update
        if key not in self.keys_down and key not in (Qt.Key.Key_Shift, Qt.Key.Key_Control, Qt.Key.Key_Alt, Qt.Key.Key_Meta):
            self.keys_pressed.add(key)
            self.keys_down.add(key)

        super().keyPressEvent(event)

    def keyReleaseEvent(self, event):
        """Handle key release events"""
        key = event.key()

        # Only handle regular keys here, modifiers are handled in timer update
        if key in self.keys_down:
            self.keys_released.add(key)
            self.keys_down.remove(key)

        super().keyReleaseEvent(event)

    def get_key(self, key: int) -> bool:
        """Check if a key is currently held down"""
        if key in (Qt.Key.Key_Shift, Qt.Key.Key_Control, Qt.Key.Key_Alt, Qt.Key.Key_Meta):
            modifier_map = {
                Qt.Key.Key_Shift: Qt.KeyboardModifier.ShiftModifier,
                Qt.Key.Key_Control: Qt.KeyboardModifier.ControlModifier,
                Qt.Key.Key_Alt: Qt.KeyboardModifier.AltModifier,
                Qt.Key.Key_Meta: Qt.KeyboardModifier.MetaModifier
            }
            return modifier_map[key] in self.modifiers_down
        return key in self.keys_down

    def get_key_down(self, key: int) -> bool:
        """Check if a key was just pressed this frame"""
        if key in (Qt.Key.Key_Shift, Qt.Key.Key_Control, Qt.Key.Key_Alt, Qt.Key.Key_Meta):
            modifier_map = {
                Qt.Key.Key_Shift: Qt.KeyboardModifier.ShiftModifier,
                Qt.Key.Key_Control: Qt.KeyboardModifier.ControlModifier,
                Qt.Key.Key_Alt: Qt.KeyboardModifier.AltModifier,
                Qt.Key.Key_Meta: Qt.KeyboardModifier.MetaModifier
            }
            return modifier_map[key] in self.modifiers_pressed
        return key in self.keys_pressed

    def get_key_up(self, key: int) -> bool:
        """Check if a key was just released this frame"""
        if key in (Qt.Key.Key_Shift, Qt.Key.Key_Control, Qt.Key.Key_Alt, Qt.Key.Key_Meta):
            modifier_map = {
                Qt.Key.Key_Shift: Qt.KeyboardModifier.ShiftModifier,
                Qt.Key.Key_Control: Qt.KeyboardModifier.ControlModifier,
                Qt.Key.Key_Alt: Qt.KeyboardModifier.AltModifier,
                Qt.Key.Key_Meta: Qt.KeyboardModifier.MetaModifier
            }
            return modifier_map[key] in self.modifiers_released
        return key in self.keys_released

    def minimumSizeHint(self):
        """Ensure the widget has a minimum size"""
        return QSize(50, 50)

    def sizeHint(self):
        """Preferred size for the widget"""
        return QSize(400, 400)

    def focusInEvent(self, event):
        """Handle focus gained"""
        super().focusInEvent(event)

    def focusOutEvent(self, event):
        """Handle focus lost"""
        super().focusOutEvent(event)
